<!doctype html>
<head>
  <title>ngshare Database Dump</title>
  <script src="masonry.min.js"></script>
  <style type="text/css">
	.grid-item {
		margin: 1em;
	}
	h2 {
		margin: 0em;
	}
  </style>
</head>
<body>
<h1>
  Database Dump for 
  <a style="color: black; text-decoration: none"
     href="https://github.com/lxylxy123456/ngshare">
    ngshare
    <img src="favicon.ico" style="height: 1em" />
  </a>
</h1>
<div data-masonry='{ "itemSelector": ".grid-item" }'>
  {% for table in tables %}
    <div class="grid-item">
      <h2>{{ table['header'] }}</h2>
      <table border="1">
        <thead>
          <tr>
            {% for cell in table['thead'] %}
              <td><b>{{ cell }}</b></td>
            {% end %}
          </tr>
        </thead>
        <tbody>
          {% for line in table['tbody'] %}
            <tr>
              {% for cell in line %}
                <td>{{ cell }}</td>
              {% end %}
            </tr>
          {% end %}
        </tbody>
      </table>
    </div>
  {% end %}
</div>
</body>
